"""
Copyright (c) 2023, Oracle and/or its affiliates.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""

from dataclasses import dataclass
from typing import Optional, Set

import agate
from dbt.adapters.oracle.relation_configs.base import OracleRelationConfigBase
from dbt.adapters.relation_configs import RelationResults, RelationConfigChange

from dbt.contracts.relation import ComponentName

from dbt.contracts.graph.nodes import ModelNode
from dbt.exceptions import DbtRuntimeError
from dbt.events import AdapterLogger

logger = AdapterLogger("oracle")

@dataclass(frozen=True, eq=True, unsafe_hash=True)
class OracleMaterializedViewConfig(OracleRelationConfigBase):
    """
    mview_name: Name of the materialized view

    query: Query that defines the materialized view

    refresh_mode: Refresh mode of the materialized view
        DEMAND - Oracle Database refreshes this materialized view whenever an appropriate refresh procedure is called
        COMMIT - Oracle Database refreshes this materialized view when a transaction on one of the materialized view's masters commits
        NEVER - Oracle Database never refreshes this materialized view

    refresh_method: Default method used to refresh the materialized view (can be overridden through the API)
        COMPLETE (C) - Materialized view is completely refreshed from the masters
        FORCE (?) - Oracle Database performs a fast refresh if possible, otherwise a complete refresh
        FAST (F) - Oracle Database performs an incremental refresh applying changes that correspond to changes in the masters since the last refresh
        NEVER (N) - User specified that the Oracle Database should not refresh this materialized view

    build_mode: Indicates how the materialized view was populated during creation
        IMMEDIATE - Populated from the masters during creation
        DEFERRED - Not populated during creation. Must be explicitly populated later by the user.
        PREBUILT - Populated with an existing table during creation. The relationship of the contents of this prebuilt table to the materialized view's masters is unknown to the Oracle Database.

    rewrite_enabled: Indicates whether rewrite is enabled (Y) or not (N)

    """
    mview_name: str
    query: str
    refresh_mode: Optional[str] = "demand"
    refresh_method: Optional[str] = "force"
    build_mode: Optional[str] = "immediate"
    query_rewrite: Optional[str] = "disable"

    @classmethod
    def from_dict(cls, config_dict) -> "OracleMaterializedViewConfig":
        kwargs_dict = {
            "mview_name": cls._render_part(ComponentName.Identifier, config_dict.get("mview_name")),
            "query": config_dict.get("query"),
            "refresh_mode": config_dict.get("refresh_mode"),
            "refresh_method": config_dict.get("refresh_method"),
            "build_mode": config_dict.get("build_mode"),
            "query_rewrite": config_dict.get("query_rewrite")
        }

        materialized_view: "OracleMaterializedViewConfig" = super().from_dict(kwargs_dict)  # type: ignore
        return materialized_view

    @classmethod
    def from_model_node(cls, model_node: ModelNode) -> "OracleMaterializedViewConfig":
        materialized_view_config = cls.parse_model_node(model_node)
        materialized_view = cls.from_dict(materialized_view_config)
        return materialized_view

    @classmethod
    def parse_model_node(cls, model_node: ModelNode) -> dict:
        config_dict = {
            "mview_name": model_node.identifier,
            "refresh_mode": model_node.config.get("refresh_mode", "demand"),
            "refresh_method": model_node.config.get("refresh_method", "force"),
            "build_mode": model_node.config.get("build_mode", "immediate"),
            "query_rewrite": model_node.config.get("query_rewrite", "disable")
        }
        if query := model_node.compiled_code:
            config_dict.update({"query": query.strip()})

        return config_dict

    @classmethod
    def from_relation_results(cls, relation_results: RelationResults) -> "OracleMaterializedViewConfig":
        materialized_view_config = cls.parse_relation_results(relation_results)
        materialized_view = cls.from_dict(materialized_view_config)
        return materialized_view

    @classmethod
    def parse_relation_results(cls, relation_results: RelationResults) -> dict:
        """
        Translate agate objects from the database into a standard dictionary.
        """

        materialized_view: agate.Row = cls._get_first_row(relation_results.get("materialized_view"))
        logger.debug(f"Materialized View data is {materialized_view}")
        config_dict = {
            "mview_name": materialized_view.get("mview_name".upper()),
            "refresh_mode": materialized_view.get("refresh_mode".upper()),
            "refresh_method": materialized_view.get("refresh_method".upper()),
            "build_mode": materialized_view.get("build_mode".upper()),
            "query": materialized_view.get("query".upper()),
        }
        logger.debug(f"Materialized View config is {config_dict}")
        if materialized_view.get('rewrite_enabled'.upper()).upper() == 'Y':
            config_dict["query_rewrite"] = "enable"
        else:
            config_dict["query_rewrite"] = "disable"

        return config_dict


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class OracleRefreshModeConfigChange(RelationConfigChange):
    context: Optional[str] = None

    @property
    def requires_full_refresh(self) -> bool:
        return False


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class OracleRefreshMethodConfigChange(RelationConfigChange):
    context: Optional[str] = None

    @property
    def requires_full_refresh(self) -> bool:
        return True

@dataclass(frozen=True, eq=True, unsafe_hash=True)
class OracleBuildModeConfigChange(RelationConfigChange):
    context: Optional[str] = None

    @property
    def requires_full_refresh(self) -> bool:
        return False

@dataclass(frozen=True, eq=True, unsafe_hash=True)
class OracleQueryRewriteConfigChange(RelationConfigChange):
    context: Optional[str] = None

    @property
    def requires_full_refresh(self) -> bool:
        return False

@dataclass(frozen=True, eq=True, unsafe_hash=True)
class OracleQueryConfigChange(RelationConfigChange):
    context: Optional[str] = None

    @property
    def requires_full_refresh(self) -> bool:
        return True


@dataclass
class OracleMaterializedViewConfigChangeset:
    refresh_mode: Optional[OracleRefreshMethodConfigChange] = None
    refresh_method: Optional[OracleRefreshMethodConfigChange] = None
    build_mode: Optional[OracleBuildModeConfigChange] = None
    query_rewrite: Optional[OracleQueryRewriteConfigChange] = None
    query: Optional[OracleQueryConfigChange] = None

    @property
    def requires_full_refresh(self) -> bool:
        return any(
            {
                self.refresh_mode.requires_full_refresh if self.refresh_mode else False,
                self.refresh_method.requires_full_refresh if self.refresh_method else False,
                self.build_mode.requires_full_refresh if self.build_mode else False,
                self.query_rewrite.requires_full_refresh if self.query_rewrite else False,
                self.query.requires_full_refresh if self.query else False
            }
        )

    @property
    def has_changes(self) -> bool:
        return any(
            {
                self.refresh_mode if self.refresh_mode else False,
                self.refresh_method if self.refresh_method else False,
                self.build_mode if self.build_mode else False,
                self.query_rewrite if self.query_rewrite else False,
                self.query if self.query else False
            }
        )



"""
Copyright (c) 2023, Oracle and/or its affiliates.
Copyright (c) 2020, Vitor Avancini

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""
import enum
import os

import dbt.exceptions
from dbt.events import AdapterLogger

from dbt.ui import warning_tag, yellow, red

logger = AdapterLogger("oracle")


class OracleNetConfig(dict):
    """The sqlnet.ora file is only supported in the python-oracledb Thick mode.
    In Thin mode, the user can pass these as environment variables

      - ssl_server_cert_dn: the distinguished name (DN) which should be
        matched with the server. This value is ignored if the
        ssl_server_dn_match parameter is not set to the value True.

      - ssl_server_dn_match: boolean indicating whether the server
        certificate distinguished name (DN) should be matched in addition to
        the regular certificate verification that is performed.

      - wallet_password: the password to use to decrypt the wallet, if it is
        encrypted. This value is only used in thin mode (default: None)

      - wallet_location: the directory where the wallet can be found. In thin
        mode this must be the directory containing the PEM-encoded wallet
        file ewallet.pem. In thick mode this must be the directory containing
        the file cwallet.sso (default: None)

      - expire_time: an integer indicating the number of minutes between the
        sending of keepalive probes. If this parameter is set to a value
        greater than zero it enables keepalive (default: 0)

      - retry_count: the number of times that a connection attempt should be
        retried before the attempt is terminated (default: 0)

      - retry_delay: the number of seconds to wait before making a new
        connection attempt (default: 0)

      - tcp_connect_timeout: a float indicating the maximum number of seconds
        to wait for establishing a connection to the database host (default:
        60.0)

      - config_dir: directory in which the optional tnsnames.ora
        configuration file is located. This value is only used in thin mode.
        For thick mode use the config_dir parameter of init_oracle_client()

    """

    keys = (
        'ssl_server_cert_dn',
        'ssl_server_dn_match',
        'wallet_password',
        'wallet_location',
        'expire_time',
        'https_proxy',
        'https_proxy_port',
        'retry_count',
        'retry_delay',
        'tcp_connect_timeout',
        'config_dir',
        'disable_oob'
    )

    @classmethod
    def from_env(cls):
        data = {}
        for key in OracleNetConfig.keys:
            val = os.getenv(key.upper())
            if val is not None:
                data[key] = val
        return cls(data)


class OracleDriverType(str, enum.Enum):
    """Database Driver Type"""
    THIN = "THIN"
    THICK = "THICK"
    CX_ORACLE = "CX"


SQLNET_ORA_CONFIG = None

# Set the environment variable ORA_PYTHON_DRIVER_TYPE to one of “cx”, “thin”, or “thick”:
ORA_PYTHON_DRIVER_TYPE = os.getenv('ORA_PYTHON_DRIVER_TYPE', 'thin').upper()
if ORA_PYTHON_DRIVER_TYPE == OracleDriverType.CX_ORACLE:
    logger.info("Running in cx mode")
    description = (
        f"cx_oracle is no longer maintained, use python-oracledb"
        f"\n\nTo switch to python-oracledb set the environment variable ORA_PYTHON_DRIVER_TYPE=thin "
        f"\n\nStarting with dbt-oracle version 1.7, default value of ORA_PYTHON_DRIVER_TYPE is thin"
        f"\n\nRead the guideline here: "
        f"https://docs.getdbt.com/reference/warehouse-setups/oracle-setup#configure-the-python-driver-mode"
        f"\n\nDocumentation for python-oracledb can be found here: "
        f"https://oracle.github.io/python-oracledb/"
    )
    logger.warning(warning_tag(red(description)))
    import cx_Oracle as oracledb
elif ORA_PYTHON_DRIVER_TYPE == OracleDriverType.THICK:
    import oracledb
    logger.info("Running in thick mode")
    oracledb.init_oracle_client()
elif ORA_PYTHON_DRIVER_TYPE == OracleDriverType.THIN:
    import oracledb
    SQLNET_ORA_CONFIG = OracleNetConfig.from_env()
    logger.info("Running in thin mode")
else:
    raise dbt.exceptions.DbtRuntimeError("Invalid value set for ORA_PYTHON_DRIVER_TYPE\n"
                                         "Use any one of 'cx', 'thin', or 'thick'")



"""
Copyright (c) 2022, Oracle and/or its affiliates.
Copyright (c) 2020, Vitor Avancini

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""

import pytest
from dbt.tests.util import run_dbt, get_relation_columns

# seeds/my_seed.csv
my_seed_csv = """
id,name,some_date
1,Easton,1981-05-20T06:46:51
2,Lillian,1978-09-03T18:10:33
3,Jeremiah,1982-03-11T03:59:51
4,Nolan,1976-05-06T20:21:35
""".lstrip()

# models/my_model.sql
my_model_sql = """
{{config(materialized='table')}}
select * from {{ ref('my_seed') }}
"""

# wrapper macro which calls the alter_column_type macro
alter_column_type_wrapper_macro = """
{% macro wrap_alter_column_type() %}
    {%- set relation = adapter.get_relation(database=target.database, schema=schema, identifier='my_model') -%}
    {{ return(adapter.dispatch('alter_column_type')(relation, 'name', 'CLOB')) }}
{% endmacro %}
"""


class TestAlterColumnDataTypeMacro:
    """
    Tests the macro `oracle__alter_column_type`

    """

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "my_seed.csv": my_seed_csv,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_sql,
        }

    @pytest.fixture(scope="class")
    def macros(self):
        return {"wrap_alter_column_type.sql": alter_column_type_wrapper_macro}

    def test_run_macro(self, project):
        """seed, then run, then macro to alter table column datatype

        """
        results = run_dbt(['seed'])
        assert len(results) == 1

        results = run_dbt(['run'])
        assert len(results) == 1

        run_dbt(['run-operation', 'wrap_alter_column_type'])
        columns = get_relation_columns(project.adapter, 'my_model')
        expected_columns = [('ID', 'NUMBER', 0),
                            ('NAME', 'CLOB', 0),
                            ('SOME_DATE', 'TIMESTAMP(6)', 0)]
        assert expected_columns == columns





"""
Copyright (c) 2022, Oracle and/or its affiliates.
Copyright (c) 2020, Vitor Avancini

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""
import pytest

from dbt.tests.adapter.utils.data_types.test_type_boolean import BaseTypeBoolean
from dbt.tests.adapter.utils.data_types.test_type_bigint import BaseTypeBigInt
from dbt.tests.adapter.utils.data_types.test_type_float import BaseTypeFloat
from dbt.tests.adapter.utils.data_types.test_type_int import BaseTypeInt
from dbt.tests.adapter.utils.data_types.test_type_numeric import BaseTypeNumeric


models__bigint_expected_sql = """
select 9223372036854775800 as bigint_col from dual
""".lstrip()

models__bigint_actual_sql = """
select cast('9223372036854775800' as {{ type_bigint() }}) as bigint_col from dual
"""

models__float_actual_sql = """
select cast('1.2345' as {{ type_float() }}) as float_col from dual
"""

models__int_actual_sql = """
select cast('12345678' as {{ type_int() }}) as int_col from dual
"""

models__numeric_actual_sql = """
select cast('1.2345' as {{ type_numeric() }}) as numeric_col from dual
"""


seeds__boolean_expected_csv = """boolean_col
1
""".lstrip()

models__boolean_actual_sql = """
select cast('1' as {{ type_boolean() }}) as boolean_col from dual
"""


class TestTypeBigIntOracle(BaseTypeBigInt):

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "expected.sql": models__bigint_expected_sql,
            "actual.sql": self.interpolate_macro_namespace(models__bigint_actual_sql, "type_bigint"),
        }


class TestTypeFloatOracle(BaseTypeFloat):

    @pytest.fixture(scope="class")
    def models(self):
        return {"actual.sql": self.interpolate_macro_namespace(models__float_actual_sql, "type_float")}


class TestTypeIntOracle(BaseTypeInt):

    @pytest.fixture(scope="class")
    def models(self):
        return {"actual.sql": self.interpolate_macro_namespace(models__int_actual_sql, "type_int")}


class TestTypeNumericOracle(BaseTypeNumeric):

    @pytest.fixture(scope="class")
    def models(self):
        return {"actual.sql": self.interpolate_macro_namespace(models__numeric_actual_sql, "type_numeric")}


class TestTypeBooleanOracle(BaseTypeBoolean):

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"expected.csv": seeds__boolean_expected_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {"actual.sql": self.interpolate_macro_namespace(models__boolean_actual_sql, "type_boolean")}




"""
Copyright (c) 2022, Oracle and/or its affiliates.
Copyright (c) 2020, Vitor Avancini

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""

import pytest
from pathlib import Path

from dbt.tests.util import run_dbt

# seeds/my_seed.csv
my_seed_csv = """
id,name,some_date
1,Easton,1981-05-20T06:46:51
2,Lillian,1978-09-03T18:10:33
3,Jeremiah,1982-03-11T03:59:51
4,Nolan,1976-05-06T20:21:35
""".lstrip()


cc_all_snapshot_sql = """
{% snapshot cc_all_snapshot %}
    {{ config(
        check_cols='all', 
        unique_key='id', 
        strategy='check',
        target_database=database, 
        target_schema=schema,
        invalidate_hard_deletes=True
    ) }}
    SELECT * FROM {{ ref('seed') }}
{% endsnapshot %}
""".strip()

# seeds/insert.sql
seeds__insert_sql = """
INSERT ALL
    INTO {schema}.seed (id, name, some_date) VALUES 
        (5, 'John Doe', TO_DATE('1982-02-03', 'YYYY-MM-DD'))
SELECT * FROM dual
"""

# seeds/update.sql
seeds__update_sql = """
UPDATE {schema}.seed 
    SET name = 'Lord Easton'
    WHERE id = 1
"""

# seeds/delete.sql
seeds__delete_sql = """
DELETE FROM {schema}.seed WHERE id = 2
"""


class TestSnapshotCheckInvalidateHardDeletes:

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": my_seed_csv,
            "insert.sql": seeds__insert_sql,
            "update.sql": seeds__update_sql,
            "delete.sql": seeds__delete_sql

        }

    @pytest.fixture(scope="class")
    def snapshots(self):
        return {
            "cc_all_snapshot.sql": cc_all_snapshot_sql,
        }

    def test_run_dbt(self, project):
        """dbt seed
           dbt snapshot
           Perform insert/update/delete
           dbt snapshot

        MERGE INTO dbt_test.cc_all_snapshot d
            USING o$pt_cc_all_snapshot182811 s
                ON (s.dbt_scd_id = d.dbt_scd_id)
            WHEN MATCHED
                THEN UPDATE
                SET dbt_valid_to = s.dbt_valid_to
                WHERE d.dbt_valid_to IS NULL
                    AND s.dbt_change_type IN ('update', 'delete')
            WHEN NOT MATCHED
                THEN INSERT (d.id, d.name, d.some_date, d.dbt_updated_at, d.dbt_valid_from, d.dbt_valid_to, d.dbt_scd_id)
                VALUES (s.id, s.name, s.some_date, s.dbt_updated_at, s.dbt_valid_from, s.dbt_valid_to, s.dbt_scd_id)
                WHERE s.dbt_change_type = 'insert'

        """
        results = run_dbt(['seed'])
        assert len(results) == 1

        # snapshot command
        results = run_dbt(["snapshot"])
        for result in results:
            assert result.status == "success"

        project.run_sql_file(Path("seeds") / Path("insert.sql"))
        project.run_sql_file(Path("seeds") / Path("update.sql"))
        project.run_sql_file(Path("seeds") / Path("delete.sql"))

        # run snapshot command
        results = run_dbt(["snapshot"])
        for result in results:
            assert result.status == "success"

        snapshot_of_updated_rows = project.run_sql(f"select * from cc_all_snapshot where id=1", fetch="all")
        assert len(snapshot_of_updated_rows) == 2

        # Deleted record will be invalidated. r['dbt_valid_to'] is set to current timestamp.
        snapshot_of_deleted_rows = project.run_sql(f"select * from cc_all_snapshot where id=2", fetch="all")
        assert len(snapshot_of_deleted_rows) == 1






"""
Copyright (c) 2023, Oracle and/or its affiliates.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""

from dataclasses import dataclass

from dbt.adapters.base.relation import Policy
from dbt.dataclass_schema import StrEnum


@dataclass
class OracleQuotePolicy(Policy):
    database: bool = False
    schema: bool = False
    identifier: bool = False


@dataclass
class OracleIncludePolicy(Policy):
    database: bool = False
    schema: bool = True
    identifier: bool = True



"""
Copyright (c) 2023, Oracle and/or its affiliates.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""
import datetime
import http
import json
from typing import Dict
import uuid
import platform

import requests
import time

import dbt.exceptions
from dbt.adapters.oracle import OracleAdapterCredentials
from dbt.events import AdapterLogger
from dbt.ui import red, green
from dbt.version import __version__ as dbt_version

# ADB-S OML Rest API minimum timeout is 1800 seconds
DEFAULT_TIMEOUT_IN_SECONDS = 1800
DEFAULT_DELAY_BETWEEN_POLL_IN_SECONDS = 2

OMLUSERS_OAUTH_API = "/omlusers/api/oauth2/v1/token"
OML_DO_EVAL_API = "/oml/api/py-scripts/v1/do-eval/{script_name}"

logger = AdapterLogger("oracle")


class OracleOML4PYClient:

    def __init__(self, oml_cloud_service_url, username, password):
        self.base_url = oml_cloud_service_url
        self._username = username
        self._password = password
        self.token = None
        self.token_expires_at = None
        self.token_url = self.base_url + OMLUSERS_OAUTH_API
        self._session = requests.Session()

    @property
    def session(self):
        return self._session

    def get_token(self):
        """Get access_token or refresh_token"""
        # If access token is about to expire then refresh the token
        if self.token_expires_at and self.token_expires_at - datetime.datetime.utcnow() < datetime.timedelta(minutes=1):
            return self._get_token(grant_type="refresh_token")
        elif self.token:  # Token is valid
            return self.token
        else:  # Generate a new token
            return self._get_token(grant_type="password")

    def _get_token(self, grant_type="password"):
        """Gets access_token or refresh_token using /broker/pdbcs/private/v1/token"""
        data = {"grant_type": grant_type}
        if grant_type == "password":
            data["username"] = self._username
            data["password"] = self._password
        else:
            data["token"] = self.token

        r = self.session.post(
            url=self.token_url,
            json=data,
            headers={
                "Accept": "application/json",
                "Content-type": "application/json",
            },
        )
        r.raise_for_status()
        response = r.json()
        self.token = response["accessToken"]
        self.token_expires_at = datetime.datetime.utcnow() + datetime.timedelta(seconds=response["expiresIn"])
        return self.token

    @property
    def default_headers(self):
        """Default headers added to every request"""
        return {
            "Content-type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {self.get_token()}",
        }

    def request(self, method: str, path: str,
                raise_for_status: bool = False,
                **kwargs) -> requests.Response:
        """
        Description:
            Perform a desired action (GET, PUT, POST) on a certain resource

        Args:
            method (str) -> HTTP verb like GET, PUT, POST, etc
            path (str) -> path to the resource e.g. /job/{job_id}
            raise_for_status (bool) -> True if HTTPError should be raised in case of an error else False

        Returns:
            object of type request.Response

        Raises:
            requests.HTTPError() in case of en error, if raise_for_status is True

        """
        url = path if path.startswith(self.base_url) else self.base_url + path
        self.session.headers.update(self.default_headers)
        r = self.session.request(method=method, url=url, **kwargs)
        try:
            r.raise_for_status()
        except requests.HTTPError:
            if raise_for_status:
                raise
        return r


class OracleADBSPythonJob:
    """Callable to submit Python Script to ADB-S

    """

    def __init__(self,
                 parsed_model: Dict,
                 credential: OracleAdapterCredentials) -> None:
        self.identifier = parsed_model["alias"]
        self.py_q_script_name = f"{self.identifier}_dbt_py_script"
        self.conda_env_name = parsed_model["config"].get("conda_env_name")
        self.timeout = parsed_model["config"].get("timeout", DEFAULT_TIMEOUT_IN_SECONDS)
        self.async_flag = parsed_model["config"].get("async_flag", False)
        self.service = parsed_model["config"].get("service", "HIGH")
        self.oml4py_client = OracleOML4PYClient(oml_cloud_service_url=credential.oml_cloud_service_url,
                                                username=credential.user,
                                                password=credential.password)
        self.session_info = self.get_session_info(credentials=credential)

    @staticmethod
    def get_session_info(credentials):
        default_action = "DBT RUN OML4PY"
        default_client_identifier = f'dbt-oracle-client-{uuid.uuid4()}'
        default_client_info = "_".join([platform.node(), platform.machine()])
        default_module = f'dbt-{dbt_version}'
        return {
            "action": credentials.session_info.get("action", default_action),
            "client_identifier": credentials.session_info.get("client_identifier", default_client_identifier),
            "clientinfo": credentials.session_info.get("client_info", default_client_info),
            "module": credentials.session_info.get("module", default_module)
        }

    def schedule_async_job_and_wait_for_completion(self, data):
        logger.info(f"Running Python aysnc job using {data}")
        try:
            r = self.oml4py_client.request(method="POST",
                                           path=OML_DO_EVAL_API.format(script_name=self.py_q_script_name),
                                           data=json.dumps(data),
                                           raise_for_status=False)
            if r.status_code in (http.HTTPStatus.BAD_REQUEST, http.HTTPStatus.INTERNAL_SERVER_ERROR):
                logger.error(red(r.json()))
            r.raise_for_status()
        except requests.exceptions.RequestException as e:
            logger.error(red(f"Error {e} scheduling async Python job for model {self.identifier}"))
            raise dbt.exceptions.DbtRuntimeError(f"Error scheduling Python model {self.identifier}")

        job_location = r.headers["location"]
        logger.info(f"Started async job {job_location}")
        start_time = time.time()

        while time.time() - start_time < self.timeout:
            logger.debug(f"Checking Job status for : {job_location}")
            try:
                job_status = self.oml4py_client.request(method="GET",
                                                        path=job_location,
                                                        raise_for_status=False)
                job_status_code = job_status.status_code
                logger.debug(f"Job status code is: {job_status_code}")
                if job_status_code == http.HTTPStatus.FOUND:
                    logger.info(green(f"Job {job_location} completed"))
                    job_result = self.oml4py_client.request(method="GET",
                                                            path=f"{job_location}/result",
                                                            raise_for_status=False)
                    job_result_json = job_result.json()
                    if 'errorMessage' in job_result_json:
                        logger.error(red(f"FAILURE - Python model {self.identifier} Job failure is: {job_result_json}"))
                        raise dbt.exceptions.DbtRuntimeError(f"Error running Python model {self.identifier}")
                    job_result.raise_for_status()
                    logger.info(green(f"SUCCESS - Python model {self.identifier} Job result is: {job_result_json}"))
                    return
                elif job_status_code == http.HTTPStatus.INTERNAL_SERVER_ERROR:
                    logger.error(red(f"FAILURE - Job status is: {job_status.json()}"))
                    raise dbt.exceptions.DbtRuntimeError(f"Error running Python model {self.identifier}")
                else:
                    logger.debug(f"Python model {self.identifier} job status is: {job_status.json()}")
                    job_status.raise_for_status()

            except requests.exceptions.RequestException as e:
                logger.error(red(f"Error {e} checking status of Python job {job_location}  for model {self.identifier}"))
                raise dbt.exceptions.DbtRuntimeError(f"Error checking status for job {job_location}")

            time.sleep(DEFAULT_DELAY_BETWEEN_POLL_IN_SECONDS)
        logger.error(red(f"Timeout error for Python model {self.identifier}"))
        raise dbt.exceptions.DbtRuntimeError(f"Timeout error for Python model {self.identifier}")

    def __call__(self, *args, **kwargs):
        data = {
            "service": self.service,
            "parameters": json.dumps(self.session_info)
        }
        if self.async_flag:
            data["asyncFlag"] = self.async_flag
            data["timeout"] = self.timeout
        if self.conda_env_name:
            data["envName"] = self.conda_env_name

        if self.async_flag:
            self.schedule_async_job_and_wait_for_completion(data=data)
        else:  # Run in blocking mode
            logger.info(f"Running Python model {self.identifier} with args {data}")
            try:
                r = self.oml4py_client.request(method="POST",
                                               path=OML_DO_EVAL_API.format(script_name=self.py_q_script_name),
                                               data=json.dumps(data),
                                               raise_for_status=False)
                job_result = r.json()
                if 'errorMessage' in job_result:
                    logger.error(red(f"FAILURE - Python model {self.identifier} Job failure is: {job_result}"))
                    raise dbt.exceptions.DbtRuntimeError(f"Error running Python model {self.identifier}")
                r.raise_for_status()
                logger.info(green(f"SUCCESS - Python model {self.identifier} Job result is: {job_result}"))
            except requests.exceptions.RequestException as e:
                logger.error(red(f"Error {e} running Python model {self.identifier}"))
                raise dbt.exceptions.DbtRuntimeError(f"Error running Python model {self.identifier}")




"""
Copyright (c) 2022, Oracle and/or its affiliates.
Copyright (c) 2020, Vitor Avancini

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""

# https://docs.oracle.com/en/database/oracle/oracle-database/21/zzpre/Oracle-reserved-words-keywords-namespaces.html#GUID-25FE5FB4-5B17-4AFA-9B59-77B6036EF579
RESERVED_KEYWORDS = {
        "ACCESS"
        "ELSE",
        "MODIFY",
        "START",
        "ADD",
        "EXCLUSIVE",
        "NOAUDIT",
        "SELECT",
        "ALL",
        "EXISTS",
        "NOCOMPRESS",
        "SESSION",
        "ALTER",
        "FILE",
        "NOT",
        "SET",
        "AND",
        "FLOAT",
        "NOTFOUND",
        "SHARE",
        "ANY",
        "FOR",
        "NOWAIT",
        "SIZE",
        "ARRAYLEN",
        "FROM",
        "NULL",
        "SMALLINT",
        "AS",
        "GRANT",
        "NUMBER",
        "SQLBUF",
        "ASC",
        "GROUP",
        "OF",
        "SUCCESSFUL",
        "AUDIT",
        "HAVING",
        "OFFLINE",
        "SYNONYM",
        "BETWEEN",
        "IDENTIFIED",
        "ON",
        "SYSDATE",
        "BY",
        "IMMEDIATE",
        "ONLINE",
        "TABLE",
        "CHAR",
        "IN",
        "OPTION",
        "THEN",
        "CHECK",
        "INCREMENT",
        "OR",
        "TO",
        "CLUSTER",
        "INDEX",
        "ORDER",
        "TRIGGER",
        "COLUMN",
        "INITIAL",
        "PCTFREE",
        "UID",
        "COMMENT",
        "INSERT",
        "PRIOR",
        "UNION",
        "COMPRESS",
        "INTEGER",
        "PRIVILEGES",
        "UNIQUE",
        "CONNECT",
        "INTERSECT",
        "PUBLIC",
        "UPDATE",
        "CREATE",
        "INTO",
        "RAW",
        "USER",
        "CURRENT",
        "IS",
        "RENAME",
        "VALIDATE",
        "DATE",
        "LEVEL",
        "RESOURCE",
        "VALUES",
        "DECIMAL",
        "LIKE",
        "REVOKE",
        "VARCHAR",
        "DEFAULT",
        "LOCK",
        "ROW",
        "VARCHAR2",
        "DELETE",
        "LONG",
        "ROWID",
        "VIEW",
        "DESC",
        "MAXEXTENTS",
        "ROWLABEL",
        "WHENEVER",
        "DISTINCT",
        "MINUS",
        "ROWNUM",
        "WHERE",
        "DROP",
        "MODE",
        "ROWS",
        "WITH"
}

NON_RESERVED_KEYWORDS = {
        "ADMIN",
        "CURSOR",
        "FOUND",
        "MOUNT",
        "AFTER",
        "CYCLE",
        "FUNCTION",
        "NEXT",
        "ALLOCATE",
        "DATABASE",
        "GO",
        "NEW",
        "ANALYZE",
        "DATAFILE",
        "GOTO",
        "NOARCHIVELOG",
        "ARCHIVE",
        "DBA",
        "GROUPS",
        "NOCACHE",
        "ARCHIVELOG",
        "DEC",
        "INCLUDING",
        "NOCYCLE",
        "AUTHORIZATION",
        "DECLARE",
        "INDICATOR",
        "NOMAXVALUE",
        "AVG",
        "DISABLE",
        "INITRANS",
        "NOMINVALUE",
        "BACKUP",
        "DISMOUNT",
        "INSTANCE",
        "NONE",
        "BEGIN",
        "DOUBLE",
        "INT",
        "NOORDER",
        "BECOME",
        "DUMP",
        "KEY",
        "NORESETLOGS",
        "BEFORE",
        "EACH",
        "LANGUAGE",
        "NORMAL",
        "BLOCK",
        "ENABLE",
        "LAYER",
        "NOSORT",
        "BODY",
        "END",
        "LINK",
        "NUMERIC",
        "CACHE",
        "ESCAPE",
        "LISTS",
        "OFF",
        "CANCEL",
        "EVENTS",
        "LOGFILE",
        "OLD",
        "CASCADE",
        "EXCEPT",
        "MANAGE",
        "ONLY",
        "CHANGE",
        "EXCEPTIONS",
        "MANUAL",
        "OPEN",
        "CHARACTER",
        "EXEC",
        "MAX",
        "OPTIMAL",
        "CHECKPOINT",
        "EXPLAIN",
        "MAXDATAFILES",
        "OWN",
        "CLOSE",
        "EXECUTE",
        "MAXINSTANCES",
        "PACKAGE",
        "COBOL",
        "EXTENT",
        "MAXLOGFILES",
        "PARALLEL",
        "COMMIT",
        "EXTERNALLY",
        "MAXLOGHISTORY",
        "PCTINCREASE",
        "COMPILE",
        "FETCH",
        "MAXLOGMEMBERS",
        "PCTUSED",
        "CONSTRAINT",
        "FLUSH",
        "MAXTRANS",
        "PLAN",
        "CONSTRAINTS",
        "FREELIST",
        "MAXVALUE",
        "PLI",
        "CONTENTS",
        "FREELISTS",
        "MIN",
        "PRECISION",
        "CONTINUE",
        "FORCE",
        "MINEXTENTS",
        "PRIMARY",
        "CONTROLFILE",
        "FOREIGN",
        "MINVALUE",
        "PRIVATE",
        "COUNT",
        "FORTRAN",
        "MODULE",
        "PROCEDURE",
        "PROFILE",
        "SAVEPOINT",
        "SQLSTATE",
        "TRACING",
        "QUOTA",
        "SCHEMA",
        "STATEMENT_ID",
        "TRANSACTION",
        "READ",
        "SCN",
        "STATISTICS",
        "TRIGGERS",
        "REAL",
        "SECTION",
        "STOP",
        "TRUNCATE",
        "RECOVER",
        "SEGMENT",
        "STORAGE",
        "UNDER",
        "REFERENCES",
        "SEQUENCE",
        "SUM",
        "UNLIMITED",
        "REFERENCING",
        "SHARED",
        "SWITCH",
        "UNTIL",
        "RESETLOGS",
        "SNAPSHOT",
        "SYSTEM",
        "USE",
        "RESTRICTED",
        "SOME",
        "TABLES",
        "USING",
        "REUSE",
        "SORT",
        "TABLESPACE",
        "WHEN",
        "ROLE",
        "SQL",
        "TEMPORARY",
        "WRITE",
        "ROLES",
        "SQLCODE",
        "THREAD",
        "WORK",
        "ROLLBACK",
        "SQLERROR",
        "TIME"
}

KEYWORDS = RESERVED_KEYWORDS | NON_RESERVED_KEYWORDS



"""
Copyright (c) 2022, Oracle and/or its affiliates.
Copyright (c) 2020, Vitor Avancini

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""

import datetime
from pathlib import Path

import pytest
from dbt.tests.util import run_dbt

# seeds/my_seed.csv
seed_csv = """
_user_id,user_name,birth_date,income,last_login_date,desc
1,Easton,1981-05-20,40000,2022-04-25T08:57:59,login
2,Lillian,1978-09-03,54000,2022-04-25T08:58:59,login
3,Jeremiah,1982-03-11,10000,2022-04-25T09:57:59,login
4,Nolan,1976-05-06,900000,2022-04-25T09:58:59,login
""".lstrip()

# models/my_incr_model.sql
my_incr_model_sql = """
{{config(materialized='incremental')}}
SELECT * FROM {{ ref('seed') }}
{% if is_incremental() %}
    WHERE last_login_date > (SELECT max(last_login_date) FROM {{ this }})
{% endif %}

"""

# seeds/add_new_rows.sql
seeds__add_new_rows_sql = """
-- insert two new rows, both of which should be in incremental model
INSERT ALL
    INTO {schema}.seed ("_user_id", user_name, birth_date, income, last_login_date, "desc") VALUES 
        (2,'Lillian Sr.', TO_DATE('1982-02-03', 'YYYY-MM-DD'), 200000, TO_DATE('2022-05-01 06:01:31', 'YYYY-MM-DD HH:MI:SS'), 'Login')
    INTO {schema}.seed ("_user_id", user_name, birth_date, income, last_login_date, "desc") VALUES 
        (5,'John Doe',TO_DATE('1992-10-01', 'YYYY-MM-DD'), 300000, TO_DATE('2022-06-01 06:01:31', 'YYYY-MM-DD HH:MI:SS'), 'Login')
SELECT * FROM dual
"""


class TestIncrementalInsertQuotedColumnsAllCols:

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": seed_csv,
            "add_new_rows.sql": seeds__add_new_rows_sql
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_incr_model.sql": my_incr_model_sql,
        }

    def test_run_dbt(self, project):
        """
            - run seed
            - run incremental model
            - add new rows
            - run incremental model

        The following SQL is expected to run.

        merge into dbt_test.my_incr_model target
          using o$pt_my_incr_model175348 temp
          on (
                    temp."_user_id" = target."_user_id"
            )
        when matched then
          update set
          target.USER_NAME = temp.USER_NAME,
          target.BIRTH_DATE = temp.BIRTH_DATE,
          target.INCOME = temp.INCOME,
          target.LAST_LOGIN_DATE = temp.LAST_LOGIN_DATE,
          target."desc" = temp."desc"
          when not matched then
          insert("_user_id", USER_NAME, BIRTH_DATE, INCOME, LAST_LOGIN_DATE, "desc")
          values(
            temp."_user_id",
            temp.USER_NAME,
            temp.BIRTH_DATE,
            temp.INCOME,
            temp.LAST_LOGIN_DATE,
            temp."desc"
            )


        """
        results = run_dbt(['seed'])
        assert len(results) == 1

        results = run_dbt(['run'])
        assert len(results) == 1

        project.run_sql_file(Path("seeds") / Path("add_new_rows.sql"))

        results = run_dbt(['run'])
        assert len(results) == 1

        user_id_2_query = 'SELECT * FROM {}.{} WHERE "_user_id" = {}'.format(project.test_schema,
                                                                             'my_incr_model',
                                                                             2)
        result = project.run_sql(user_id_2_query, fetch="all")
        assert len(result) == 2

        used_id_5_query = 'SELECT * FROM {}.{} WHERE "_user_id" = {}'.format(project.test_schema,
                                                                             'my_incr_model',
                                                                             5)
        expected_result = [(5,  'John Doe',
                            datetime.datetime(1992, 10, 1, 0, 0),
                            300000,
                            datetime.datetime(2022, 6, 1, 6, 1, 31),
                            'Login')]

        result = project.run_sql(used_id_5_query, fetch="all")
        assert result == expected_result


